indexing
	description: "INTEGER-only routines for specialized numeric processing"
	license: "Eiffel Forumn License v2 - See license.txt"
	source: "{
		Amalasoft Corporation
		273 Harwood Avenue
		Littleton, MA 01460 USA
		}"
	status: "Copyright 1995-2009, Amalasoft"
	date: "$Date: 2009/02/28 $"
	revision: "$Revision: 001$"
	howto: "{
    Instantiate using default_create, or inherit
	 }"

class AEL_NUMERIC_INTEGER_ROUTINES

create
	default_create

--|========================================================================
feature -- Conversion
--|========================================================================

	rounded_to_next_even_mod (v, m: INTEGER; zok: BOOLEAN): INTEGER is
			-- The given number 'v', rounded toward higher magnitude,
			-- to an even multiple of 'm'
			-- Positive numbers are rounded up
			-- Negative numbers are rounded down
			-- If zero is acceptable (the 'zok' flag = True) then
			-- a value of zero will yield a result of zero.  If not,
			-- then a value of zero will yield a result of 'm'
		require
			valid_mod: m > 1
		do
			Result := rounded_to_next_64_even_mod (v, m, zok).to_integer_32
		ensure
			sign_preserved: (v >= 0) = (Result >= 0)
			positive_on_request: (v = 0 and not zok) implies Result = m
			at_least: Result.abs >= v.abs
			even_mod: (Result \\ m) = 0
		end

	--|--------------------------------------------------------------

	rounded_to_next_64_even_mod (v, m: INTEGER_64; zok: BOOLEAN): INTEGER_64 is
			-- The given number 'v', rounded toward higher magnitude,
			-- to an even multiple of 'm'
			-- Positive numbers are rounded up
			-- Negative numbers are rounded down
			-- If zero is acceptable (the 'zok' flag = True) then
			-- a value of zero will yield a result of zero.  If not,
			-- then a value of zero will yield a result of 'm'
		require
			valid_mod: m > 1
		local
			tv: INTEGER_64
			is_neg: BOOLEAN
		do
			if v = 0 then
				if not zok then
					Result := m
				end
			elseif v \\ m = 0 then
				Result := v
			else
				if v < 0 then
					tv := -v
					is_neg := True
				else
					tv := v
				end
				Result := tv + (m - (tv \\ m))
				if is_neg then
					Result := -Result
				end
			end
		ensure
			sign_preserved: (v >= 0) = (Result >= 0)
			positive_on_request: (v = 0 and not zok) implies Result = m
			at_least: Result.abs >= v.abs
			even_mod: (Result \\ m) = 0
		end

	--|--------------------------------------------------------------

	even_8_multiple (v: INTEGER): INTEGER is
			-- Nearest value at or greater than 'v' that is an even
			-- multiple of 8
		require
			non_negative: v >= 0
		do
			Result := c_even_8_multiple (v)
		end

	--|--------------------------------------------------------------

	even_10_multiple (v: INTEGER): INTEGER is
			-- Nearest value at or greater than 'v' that is an even
			-- multiple of 10
		require
			non_negative: v >= 0
		do
			Result := c_even_10_multiple (v)
		end

	--|--------------------------------------------------------------

	even_8_multiple_64 (v: INTEGER_64): INTEGER_64 is
			-- Nearest 64 bit value at or greater than 'v' that is an even
			-- multiple of 8
		require
			non_negative: v >= 0
		do
			Result := rounded_to_next_64_even_mod (v, 8, False)
		end

	--|--------------------------------------------------------------

	even_10_multiple_64 (v: INTEGER_64): INTEGER_64 is
			-- Nearest 64 bit value at or greater than 'v' that is an even
			-- multiple of 10
		require
			non_negative: v >= 0
		do
			Result := rounded_to_next_64_even_mod (v, 10, False)
		end

--|========================================================================
feature -- Byte operations
--|========================================================================

	highest_bit_set (v: NATURAL_8): INTEGER is
			-- The number (from 0 to 7, of the most significant bit
			-- that is set in the given byte
			-- If none is set, then -1
		do
			if v = 0 then
				Result := -1
			else
				Result := hi_bits_by_value.item (v)
			end
		end

	--|--------------------------------------------------------------

	highest_bit_unset (v: NATURAL_8): INTEGER is
			-- The number (from 0 to 7, of the most significant bit
			-- that is UNSET in the given byte
			-- If none is UNSET (all are set), then -1
		do
			Result := highest_bit_set (byte_complement (v))
		end

	--|--------------------------------------------------------------

	lowest_bit_set (v: NATURAL_8): INTEGER is
			-- The number (from 0 to 7, of the least significant bit
			-- that is set in the given byte
			-- If none is set, then -1
			--
			-- All odd numbers have 1 as lowest bit set
		local
			neg: NATURAL_8
			bv: INTEGER
		do
			Result := -1
			if v /= 0 then
				neg := (0 - v).to_natural_8
				bv := v & neg
				inspect bv
				when 1, 2 then
					Result := bv - 1
				when 4 then
					Result := 2
				when 8 then
					Result := 3
				when 16 then
					Result := 4
				when 32 then
					Result := 5
				when 64 then
					Result := 6
				when 128 then
					Result := 7
				else
				end
			end
		end

	--|--------------------------------------------------------------

	lowest_bit_unset (v: NATURAL_8): INTEGER is
			-- The number (from 0 to 7, of the least significant bit
			-- that is UNSET in the given byte
			-- If none is UNSET (all are set), then -1
			--
			-- All even numbers have 0 as lowest bit unset
		do
			if v = 0 or v \\ 2 = 0 then
				Result := 0
			elseif v = 0xff then
				Result := -1
			else
				Result := lowest_bit_set (byte_complement (v))
			end
		end

	--|--------------------------------------------------------------

	byte_complement (v: NATURAL_8): NATURAL_8 is
			-- Complement of 'v'
		do
			Result := c_byte_complement (v)
		end

--|========================================================================
feature {NONE} -- Constants
--|========================================================================

	hi_bits_by_value: ARRAY [INTEGER] is
			-- The highest set 0-based bit positions for a given value
			-- If no bits are set, the result is -1
			--
			-- NOTE WELL that this is a 0-based array!!
		local
			i: INTEGER
		once
			create Result.make (0, 255)
			Result.put (-1, 0) -- 00000000
			Result.put (0, 1)  -- 00000001
			Result.put (1, 2)  -- 00000010
			Result.put (1, 3)  -- 00000011
			Result.put (2, 4)  -- 00000100
			Result.put (2, 5)  -- 00000101
			Result.put (2, 6)  -- 00000110
			Result.put (2, 7)  -- 00000111
			from i := 8
			until i > 15
			loop
				Result.put (3, i) -- 00001000 - 00001111
				i := i + 1
			end
			from i := 16
			until i > 31
			loop
				Result.put (4, i) -- 00010000 - 00011111
				i := i + 1
			end
			from i := 32
			until i > 63
			loop
				Result.put (5, i) -- 00100000 - 00111111
				i := i + 1
			end
			from i := 64
			until i > 127
			loop
				Result.put (6, i) -- 01000000 - 01111111
				i := i + 1
			end
			from i := 128
			until i > 255
			loop
				Result.put (7, i) -- 10000000 - 11111111
				i := i + 1
			end
		ensure
			exists: Result /= Void
			valid_range: Result.lower = 0 and Result.upper = 255
		end

--|========================================================================
feature {NONE} -- Externals
--|========================================================================

	c_byte_complement (v: NATURAL_8): NATURAL_8 is
			-- Complement of 'v'
		require
			in_range: v >= 0 and v <= 255
		external
			"C inline use %"stdlib.h%""
		alias
			"[
			{
				return ~($v);
			}
			]"
		end

	--|--------------------------------------------------------------

	c_even_8_multiple (v: INTEGER): INTEGER is
		-- Nearest value at or greater than 'v' that is an even
		-- multiple of 8
		require
			non_negative: v >= 0
		external
			"C inline use %"stdlib.h%""
		alias
			"[
			{
				int Result=8;
				if ($v != 0) {
					if (($v % 8) == 0)
						Result = $v;
					else
						Result = $v + (8 - ($v % 8));
				}
				return Result;
			}
			]"
		end

	--|--------------------------------------------------------------

	c_even_10_multiple (v: INTEGER): INTEGER is
		-- Nearest value at or greater than 'v' that is an even
		-- multiple of 10
		require
			non_negative: v >= 0
		external
			"C inline use %"stdlib.h%""
		alias
			"[
			{
				int Result=10;
				if ($v != 0) {
					if (($v % 10) == 0)
						Result = $v;
					else
						Result = $v + (10 - ($v % 10));
				}
				return Result;
			}
			]"
		end

	--|--------------------------------------------------------------

	num_bits_set_8 (byte: NATURAL_8): INTEGER is
			-- Number of bits set in the given byte
			-- Uses Kernighan's algorithm
		external
			"C inline use %"stdlib.h%""
		alias
			"[
				{
				unsigned char v; // count the number of bits set in v
				int Result=0; // c accumulates the total bits set in v
				v = $byte;
				for (; v; Result++) {
				  v &= v - 1; // clear the least significant bit set
				}
				return Result;
				}
			]"
		end

	num_bits_set_32 (num: NATURAL_32): INTEGER is
			-- Number of bits set in the given 32 bit number
			-- Uses Kernighan's algorithm
		external
			"C inline use %"stdlib.h%""
		alias
			"[
				{
				unsigned int v; // count the number of bits set in v
				int Result=0; // c accumulates the total bits set in v
				v = $num;
				for (; v; Result++) {
				  v &= v - 1; // clear the least significant bit set
				}
				return Result;
				}
			]"
		end

end -- class AEL_NUMERIC_INTEGER_ROUTINES
